package net.deterlab.abac;

import org.apache.commons.collections15.*;
import org.apache.commons.collections15.functors.*;

import edu.uci.ics.jung.graph.*;
import edu.uci.ics.jung.graph.event.*;
import edu.uci.ics.jung.graph.util.*;

import java.awt.geom.Point2D;
import java.io.*;
import java.util.*;

/**
 * Represents a global graph of credentials in the form of principals and
 * attributes.
 */
public class CredentialGraph {
    private Graph<Role,Credential> g = null;
    private java.util.Set<Credential> derived_edges = new HashSet<Credential>();
    private Query pq;

    public CredentialGraph() {
        /* create the graph */
        g = Graphs.<Role,Credential>synchronizedDirectedGraph(new DirectedSparseGraph<Role,Credential>());
        pq = new Query(g);
    }

    /**
     * Returns a Query object which can be used to query the graph.
     */
    public Query querier() {
        derive_linked_edges();
        return new Query(g);
    }

    /**
     * Add a credential to the graph.
     */
    public void add_credential(Credential cred) {
        Role tail = cred.tail();
        Role head = cred.head();

        /* explicitly add the vertices, otherwise get a null pointer exception */
        g.addVertex(tail);
        g.addVertex(head);

        g.addEdge(cred, tail, head);
    }

    /**
     * Remove a credential from the graph.
     */
    public void remove_credential(Credential cred) {
        g.removeEdge(cred);
    }

    /**
     * Returns a collection of the credentials in the graph.
     */
    public Collection<Credential> credentials() {
        Collection<Credential> creds = new HashSet<Credential>();
        for (Credential cred : g.getEdges()) {
            // FIXME: this should check if there's a cert on the cred
            // this supports all types of derived edges, including intersection

            Role head = g.getDest(cred);

            /* any edge to a linking node is derived, so skip it */
            if (!head.is_linking())
                creds.add(cred);
        }
        return creds;
    }

    /**
     * Derive the linked edges in the graph, according to RT0 derivation rules.
     * They are added to this graph and to any filtered graphs that match. See
     * "Distributed Credential Chain Discovery in Trust Management" by Ninghui
     * Li et al. for details.
     */
    private void derive_linked_edges() {
        clear_old_edges();

        /* for every node in the graph.. */
        for (Role A_r1_r2 : g.getVertices()) {
            if (!A_r1_r2.is_linking()) continue;

            Role A_r1 = new Role(A_r1_r2.A_r1());
            String r2 = A_r1_r2.r2();

            /* locate the node A.r1 */
            if (!g.containsVertex(A_r1)) continue; /* boring: nothing of the form A.r1 */

            /* for each B that satisfies A_r1 */
            for (Role principal : pq.find_principals(A_r1)) {
                Role B_r2 = new Role(principal + "." + r2);
                if (!g.containsVertex(B_r2)) continue;

                Credential derived_edge = new Credential(A_r1_r2, B_r2);
                derived_edges.add(derived_edge);
                g.addEdge(derived_edge, B_r2, A_r1_r2);
            }
        }
    }

    /**
     * Clear the derived edges that currently exist in the graph. This is done
     * before the edges are rederived. The derived edges in filtered graphs are
     * also cleared.
     */
    private void clear_old_edges() {
        for (Credential i: derived_edges)
            g.removeEdge(i);
        derived_edges = new HashSet<Credential>();
    }
}
