package net.deterlab.abac;

/**
 * Represents a role, which is a vertex in a Credential graph.
 */
public class Role {
    private String m_string;
    private String[] m_parts;
    private String m_A_r1, m_r2;
    private String m_prefix;

    /**
     * Create a role from a string. The string must be of the format "A",
     * "A.r1", or "A.r1.r2", where A is a principal and r1 and r2 are role
     * names. If the string does not have this format, the constructor throws
     * a RuntimeException.
     */
    public Role(String s) throws RuntimeException {
        m_string = s;

        m_parts = s.split("\\.");
        if (m_parts.length > 3)
            throw new RuntimeException("Not a valid role: " + m_string);

        // linking role: prefix is A.r1 from A.r1.r2
        if (is_linking()) {
            m_A_r1 = m_parts[0] + "." + m_parts[1];
            m_r2 = m_parts[2];
            m_prefix = m_A_r1;
        }

        // role: prefix is A from A.r1
        else if (is_role())
            m_prefix = m_parts[0];

        // principal: prefix is the whole thing
        else
            m_prefix = m_string;
    }

    /**
     * Returns true iff the role is a principal.
     */
    public boolean is_principal() { return m_parts.length == 1; }

    /**
     * Returns true iff the role is a role (i.e., A.r1).
     */
    public boolean is_role() { return m_parts.length == 2; }

    /**
     * Returns true iff the role is a linking role (i.e., A.r1.r2).
     */
    public boolean is_linking() { return m_parts.length == 3; }

    /**
     * Returns the first two elements of a linking role's name. This typically
     * refers to another role in the graph. This will throw a runtime
     * exception if the node is not a linking role.
     */
    public String A_r1() throws RuntimeException {
        if (!is_linking())
            throw new RuntimeException("Not a linking role");
        return m_A_r1;
    }

    /**
     * Return the last element of a linking role's name. This will throw a
     * runtime exception if the node is not a linking role.
     */
    public String r2() throws RuntimeException {
        if (!is_linking())
            throw new RuntimeException("Not a linking role");
        return m_r2;
    }

    /**
     * Returns the principal part of a role or principal. This is everything
     * except the last element of the name.
     */
    public String principal_part() {
        return m_prefix;
    }

    /**
     * Returns true if the principal part of the name matches a prefix. This
     * is used when filtering graphs.
     */
    public boolean matches(String prefix) {
        return prefix.length() == 0 || m_prefix.equals(prefix);
    }

    public String toString() {
        return m_string;
    }

    public boolean equals(Object v2) {
        if (v2 instanceof Role)
            return m_string.equals(((Role)v2).m_string);
        return false;
    }

    public int hashCode() {
        return m_string.hashCode();
    }
}
